@@ -10,10 +10,13 @@
 #include <sched.h>
 #include <setjmp.h>
 #include <signal.h>
+#include <sys/prctl.h>
 #include <sys/socket.h>
 #include <sys/syscall.h>
 #include <unistd.h>
 
+#include <utility>
+
 #include "LinuxCapabilities.h"
 #include "LinuxSched.h"
 #include "SandboxChrootProto.h"
@@ -25,12 +28,14 @@
 #include "mozilla/ArrayUtils.h"
 #include "mozilla/Assertions.h"
 #include "mozilla/Attributes.h"
-#include "mozilla/Move.h"
 #include "mozilla/Preferences.h"
 #include "mozilla/SandboxReporter.h"
 #include "mozilla/SandboxSettings.h"
-#include "mozilla/Services.h"
+#include "mozilla/Components.h"
+#include "mozilla/StaticPrefs_media.h"
+#include "mozilla/StaticPrefs_security.h"
 #include "mozilla/Unused.h"
+#include "mozilla/ipc/UtilityProcessSandboxing.h"
 #include "nsCOMPtr.h"
 #include "nsDebug.h"
 #include "nsIGfxInfo.h"
@@ -38,17 +43,26 @@
 #include "nsThreadUtils.h"
 #include "prenv.h"
 #include "sandbox/linux/system_headers/linux_syscalls.h"
+#include "sandbox/linux/services/syscall_wrappers.h"
+
+#include "mozilla/pthread_atfork.h"
 
 #ifdef MOZ_X11
 #  ifndef MOZ_WIDGET_GTK
 #    error "Unknown toolkit"
 #  endif
+#  include "mozilla/WidgetUtilsGtk.h"
 #  include <gdk/gdk.h>
 #  include <gdk/gdkx.h>
 #  include "X11UndefineNone.h"
 #  include "gfxPlatform.h"
 #endif
 
+#if defined(__GLIBC__) && !defined(__UCLIBC__)
+// We really are using glibc, not uClibc pretending to be glibc.
+#  define LIBC_GLIBC 1
+#endif
+
 namespace mozilla {
 
 // Returns true if graphics will work from a content process
@@ -59,20 +73,27 @@ namespace mozilla {
 //
 // (Longer-term we intend to either proxy or remove X11 access from
 // content processes, at which point this will stop being an issue.)
-static bool IsDisplayLocal() {
+static bool IsGraphicsOkWithoutNetwork() {
   // For X11, check whether the parent's connection is a Unix-domain
   // socket.  This is done instead of trying to parse the display name
   // because an empty hostname (e.g., ":0") will fall back to TCP in
   // case of failure to connect using Unix-domain sockets.
 #ifdef MOZ_X11
   // First, ensure that the parent process's graphics are initialized.
-  Unused << gfxPlatform::GetPlatform();
+  DebugOnly<gfxPlatform*> gfxPlatform = gfxPlatform::GetPlatform();
 
   const auto display = gdk_display_get_default();
-  if (NS_WARN_IF(display == nullptr)) {
-    return false;
+  if (!display) {
+    // In this case, the browser is headless, but WebGL could still
+    // try to use X11.  However, WebGL isn't supported with remote
+    // X11, and in any case these connections are made after sandbox
+    // startup (lazily when WebGL is used), so they aren't being done
+    // directly by the process anyway.  (For local X11, they're
+    // brokered.)
+    MOZ_ASSERT(gfxPlatform->IsHeadless());
+    return true;
   }
-  if (GDK_IS_X11_DISPLAY(display)) {
+  if (mozilla::widget::GdkIsX11Display(display)) {
     const int xSocketFd = ConnectionNumber(GDK_DISPLAY_XDISPLAY(display));
     if (NS_WARN_IF(xSocketFd < 0)) {
       return false;
@@ -100,14 +121,28 @@ static bool IsDisplayLocal() {
     //
     // Unfortunately, the Xorg client libraries prefer the abstract
     // addresses, so this isn't directly detectable by inspecting the
-    // parent process's socket.  Instead, this checks for the
-    // directory the sockets are stored in, which typically won't
-    // exist in a container with a private /tmp that isn't running its
-    // own X server.
-    if (access("/tmp/.X11-unix", X_OK) != 0) {
-      NS_ERROR(
-          "/tmp/.X11-unix is inaccessible; can't isolate network"
-          " namespace in content processes");
+    // parent process's socket.  Instead, parse the DISPLAY env var
+    // (which was updated if necessary in nsAppRunner.cpp) to get the
+    // display number and construct the socket path, falling back to
+    // testing the directory in case that doesn't work.  (See bug
+    // 1565972 and bug 1559368 for cases where we need to test the
+    // specific socket.)
+    const char* const displayStr = PR_GetEnv("DISPLAY");
+    nsAutoCString socketPath("/tmp/.X11-unix");
+    int accessFlags = X_OK;
+    int displayNum;
+    // sscanf ignores trailing text, so display names with a screen
+    // number (e.g., ":0.2") will parse correctly.
+    if (displayStr && (sscanf(displayStr, ":%d", &displayNum) == 1 ||
+                       sscanf(displayStr, "unix:%d", &displayNum) == 1)) {
+      socketPath.AppendPrintf("/X%d", displayNum);
+      accessFlags = R_OK | W_OK;
+    }
+    if (access(socketPath.get(), accessFlags) != 0) {
+      SANDBOX_LOG_ERRNO(
+          "%s is inaccessible; can't isolate network namespace in"
+          " content processes",
+          socketPath.get());
       return false;
     }
   }
@@ -119,7 +154,7 @@ static bool IsDisplayLocal() {
 }
 
 bool HasAtiDrivers() {
-  nsCOMPtr<nsIGfxInfo> gfxInfo = services::GetGfxInfo();
+  nsCOMPtr<nsIGfxInfo> gfxInfo = components::GfxInfo::Service();
   nsAutoString vendorID;
   static const Array<nsresult (nsIGfxInfo::*)(nsAString&), 2> kMethods = {
       &nsIGfxInfo::GetAdapterVendorID,
@@ -148,19 +183,21 @@ static bool ContentNeedsSysVIPC() {
   // The ALSA dmix plugin uses SysV semaphores and shared memory to
   // coordinate software mixing.
 #ifdef MOZ_ALSA
-  if (!Preferences::GetBool("media.cubeb.sandbox")) {
+  if (!StaticPrefs::media_cubeb_sandbox()) {
     return true;
   }
 #endif
 
-  // Bug 1438391: VirtualGL uses SysV shm for images and configuration.
-  if (PR_GetEnv("VGL_ISACTIVE") != nullptr) {
-    return true;
-  }
+  if (GetEffectiveContentSandboxLevel() < 5) {
+    // Bug 1438391: VirtualGL uses SysV shm for images and configuration.
+    if (PR_GetEnv("VGL_ISACTIVE") != nullptr) {
+      return true;
+    }
 
-  // The fglrx (ATI Catalyst) GPU drivers use SysV IPC.
-  if (HasAtiDrivers()) {
-    return true;
+    // The fglrx (ATI Catalyst) GPU drivers use SysV IPC.
+    if (HasAtiDrivers()) {
+      return true;
+    }
   }
 
   return false;
@@ -185,38 +222,44 @@ static void PreloadSandboxLib(base::environment_map* aEnv) {
   (*aEnv)["LD_PRELOAD"] = preload.get();
 }
 
-static void AttachSandboxReporter(base::file_handle_mapping_vector* aFdMap) {
-  int srcFd, dstFd;
-  SandboxReporter::Singleton()->GetClientFileDescriptorMapping(&srcFd, &dstFd);
-  aFdMap->push_back({srcFd, dstFd});
+static bool AttachSandboxReporter(geckoargs::ChildProcessArgs& aExtraOpts) {
+  UniqueFileHandle clientFileDescriptor(
+      dup(SandboxReporter::Singleton()->GetClientFileDescriptor()));
+  if (!clientFileDescriptor) {
+    SANDBOX_LOG_ERRNO("dup");
+    return false;
+  }
+
+  geckoargs::sSandboxReporter.Put(std::move(clientFileDescriptor), aExtraOpts);
+  return true;
 }
 
-class SandboxFork : public base::LaunchOptions::ForkDelegate {
- public:
-  explicit SandboxFork(int aFlags, bool aChroot);
-  virtual ~SandboxFork();
+static bool AttachSandboxChroot(geckoargs::ChildProcessArgs& aExtraOpts,
+                                base::LaunchOptions* aOptions) {
+  int fds[2];
+  int rv = socketpair(AF_UNIX, SOCK_STREAM | SOCK_CLOEXEC, 0, fds);
+  if (rv != 0) {
+    SANDBOX_LOG_ERRNO("socketpair");
+    return false;
+  }
 
-  void PrepareMapping(base::file_handle_mapping_vector* aMap);
-  pid_t Fork() override;
+  geckoargs::sChrootClient.Put(UniqueFileHandle{fds[0]}, aExtraOpts);
+  aOptions->sandbox_chroot_server.reset(fds[1]);
+  return true;
+}
 
- private:
-  int mFlags;
-  int mChrootServer;
-  int mChrootClient;
-
-  void StartChrootServer();
-  SandboxFork(const SandboxFork&) = delete;
-  SandboxFork& operator=(const SandboxFork&) = delete;
-};
-
-static int GetEffectiveSandboxLevel(GeckoProcessType aType) {
+static int GetEffectiveSandboxLevel(GeckoProcessType aType,
+                                    ipc::SandboxingKind aKind) {
   auto info = SandboxInfo::Get();
   switch (aType) {
-    case GeckoProcessType_GMPlugin:
-      if (info.Test(SandboxInfo::kEnabledForMedia)) {
-        return 1;
-      }
-      return 0;
+#ifdef MOZ_ENABLE_FORKSERVER
+      // With this mozsandbox will be preloaded for the fork server.  Sandboxed
+      // child processes rely on wrappers defined by mozsandbox to work
+      // properly.
+    case GeckoProcessType_ForkServer:
+      return 1;
+      break;
+#endif
     case GeckoProcessType_Content:
       // GetEffectiveContentSandboxLevel is main-thread-only due to prefs.
       MOZ_ASSERT(NS_IsMainThread());
@@ -224,34 +267,49 @@ static int GetEffectiveSandboxLevel(GeckoProcessType aType) {
         return GetEffectiveContentSandboxLevel();
       }
       return 0;
+    case GeckoProcessType_GMPlugin:
+      if (info.Test(SandboxInfo::kEnabledForMedia)) {
+        return 1;
+      }
+      return 0;
     case GeckoProcessType_RDD:
       return PR_GetEnv("MOZ_DISABLE_RDD_SANDBOX") == nullptr ? 1 : 0;
+    case GeckoProcessType_Socket:
+      // GetEffectiveSocketProcessSandboxLevel is main-thread-only due to prefs.
+      MOZ_ASSERT(NS_IsMainThread());
+      return GetEffectiveSocketProcessSandboxLevel();
+    case GeckoProcessType_Utility:
+      return IsUtilitySandboxEnabled(aKind);
     default:
       return 0;
   }
 }
 
-void SandboxLaunchPrepare(GeckoProcessType aType,
-                          base::LaunchOptions* aOptions) {
+// static
+bool SandboxLaunch::Configure(GeckoProcessType aType, SandboxingKind aKind,
+                              geckoargs::ChildProcessArgs& aExtraOpts,
+                              LaunchOptions* aOptions) {
+  MOZ_ASSERT(aOptions->fork_flags == 0 && !aOptions->sandbox_chroot_server);
   auto info = SandboxInfo::Get();
 
   // We won't try any kind of sandboxing without seccomp-bpf.
   if (!info.Test(SandboxInfo::kHasSeccompBPF)) {
-    return;
+    return true;
   }
 
   // Check prefs (and env vars) controlling sandbox use.
-  int level = GetEffectiveSandboxLevel(aType);
+  int level = GetEffectiveSandboxLevel(aType, aKind);
   if (level == 0) {
-    return;
+    return true;
   }
 
   // At this point, we know we'll be using sandboxing; generic
-  // sandboxing support goes here.  The MOZ_SANDBOXED env var tells
-  // the child process whether this is the case.
-  aOptions->env_map["MOZ_SANDBOXED"] = "1";
+  // sandboxing support goes here.
   PreloadSandboxLib(&aOptions->env_map);
-  AttachSandboxReporter(&aOptions->fds_to_remap);
+  if (aType != GeckoProcessType_ForkServer &&
+      !AttachSandboxReporter(aExtraOpts)) {
+    return false;
+  }
 
   bool canChroot = false;
   int flags = 0;
@@ -265,29 +323,52 @@ void SandboxLaunchPrepare(GeckoProcessType aType,
     } else {
       flags |= CLONE_NEWIPC;
     }
+
+    // The intent of level 5 is to block display server access, so
+    // tell the content process not to attempt to connect.
+    if (GetEffectiveContentSandboxLevel() >= 5) {
+      aOptions->env_map["MOZ_HEADLESS"] = "1";
+    }
   }
 
   // Anything below this requires unprivileged user namespaces.
   if (!info.Test(SandboxInfo::kHasUserNamespaces)) {
-    return;
+    return true;
   }
 
   switch (aType) {
+    case GeckoProcessType_Socket:
+      if (level >= 1) {
+        canChroot = true;
+        flags |= CLONE_NEWIPC;
+      }
+      break;
     case GeckoProcessType_GMPlugin:
     case GeckoProcessType_RDD:
       if (level >= 1) {
         canChroot = true;
-        flags |= CLONE_NEWNET | CLONE_NEWIPC;
+        // Can't use CLONE_NEWIPC because of intel-media-driver.
+        flags |= CLONE_NEWNET;
+      }
+      break;
+    case GeckoProcessType_Utility:
+      if (level >= 1) {
+        canChroot = true;
+        flags |= CLONE_NEWIPC;
+        flags |= CLONE_NEWNET;
       }
       break;
     case GeckoProcessType_Content:
       if (level >= 4) {
         canChroot = true;
+
         // Unshare network namespace if allowed by graphics; see
         // function definition above for details.  (The display
         // local-ness is cached because it won't change.)
         static const bool canCloneNet =
-            IsDisplayLocal() && !PR_GetEnv("RENDERDOC_CAPTUREOPTS");
+            level >= 5 || (IsGraphicsOkWithoutNetwork() &&
+                           !PR_GetEnv("RENDERDOC_CAPTUREOPTS"));
+
         if (canCloneNet) {
           flags |= CLONE_NEWNET;
         }
@@ -304,52 +385,44 @@ void SandboxLaunchPrepare(GeckoProcessType aType,
       break;
   }
 
+  if (canChroot && !AttachSandboxChroot(aExtraOpts, aOptions)) {
+    return false;
+  }
+
   if (canChroot || flags != 0) {
-    auto forker = MakeUnique<SandboxFork>(flags | CLONE_NEWUSER, canChroot);
-    forker->PrepareMapping(&aOptions->fds_to_remap);
-    aOptions->fork_delegate = std::move(forker);
-    if (canChroot) {
-      aOptions->env_map[kSandboxChrootEnvFlag] = "1";
-    }
+    flags |= CLONE_NEWUSER;
   }
+
+  aOptions->env_map[kSandboxChrootEnvFlag] = std::to_string(canChroot ? 1 : 0);
+
+  aOptions->fork_flags = flags;
+  return true;
 }
 
-SandboxFork::SandboxFork(int aFlags, bool aChroot)
-    : mFlags(aFlags), mChrootServer(-1), mChrootClient(-1) {
-  if (aChroot) {
-    int fds[2];
-    int rv = socketpair(AF_UNIX, SOCK_STREAM | SOCK_CLOEXEC, 0, fds);
-    if (rv != 0) {
-      SANDBOX_LOG_ERROR("socketpair: %s", strerror(errno));
-      MOZ_CRASH("socketpair failed");
-    }
-    mChrootClient = fds[0];
-    mChrootServer = fds[1];
-  }
-}
+SandboxLaunch::SandboxLaunch() : mFlags(0), mChrootServer(-1) {}
 
-void SandboxFork::PrepareMapping(base::file_handle_mapping_vector* aMap) {
-  if (mChrootClient >= 0) {
-    aMap->push_back({mChrootClient, kSandboxChrootClientFd});
-  }
-}
-
-SandboxFork::~SandboxFork() {
-  if (mChrootClient >= 0) {
-    close(mChrootClient);
-  }
+SandboxLaunch::~SandboxLaunch() {
   if (mChrootServer >= 0) {
     close(mChrootServer);
   }
 }
 
+bool SandboxLaunch::Prepare(LaunchOptions* aOptions) {
+  MOZ_ASSERT(mChrootServer < 0);
+
+  mFlags = aOptions->fork_flags;
+  mChrootServer = aOptions->sandbox_chroot_server.release();
+
+  return true;
+}
+
 static void BlockAllSignals(sigset_t* aOldSigs) {
   sigset_t allSigs;
   int rv = sigfillset(&allSigs);
   MOZ_RELEASE_ASSERT(rv == 0);
   rv = pthread_sigmask(SIG_BLOCK, &allSigs, aOldSigs);
   if (rv != 0) {
-    SANDBOX_LOG_ERROR("pthread_sigmask (block all): %s", strerror(rv));
+    SANDBOX_LOG_WITH_ERROR(rv, "pthread_sigmask (block all)");
     MOZ_CRASH("pthread_sigmask");
   }
 }
@@ -360,13 +433,28 @@ static void RestoreSignals(const sigset_t* aOldSigs) {
   // state right now:
   int rv = pthread_sigmask(SIG_SETMASK, aOldSigs, nullptr);
   if (rv != 0) {
-    SANDBOX_LOG_ERROR("pthread_sigmask (restore): %s", strerror(-rv));
+    SANDBOX_LOG_WITH_ERROR(rv, "pthread_sigmask (restore)");
     MOZ_CRASH("pthread_sigmask");
   }
 }
 
+static bool IsSignalIgnored(int aSig) {
+  struct sigaction sa{};
+
+  if (sigaction(aSig, nullptr, &sa) != 0) {
+    if (errno != EINVAL) {
+      SANDBOX_LOG_ERRNO("sigaction(%d)", aSig);
+    }
+    return false;
+  }
+  return sa.sa_handler == SIG_IGN;
+}
+
 static void ResetSignalHandlers() {
   for (int signum = 1; signum <= SIGRTMAX; ++signum) {
+    if (IsSignalIgnored(signum)) {
+      continue;
+    }
     if (signal(signum, SIG_DFL) == SIG_ERR) {
       MOZ_DIAGNOSTIC_ASSERT(errno == EINVAL);
     }
@@ -375,6 +463,22 @@ static void ResetSignalHandlers() {
 
 namespace {
 
+#if defined(LIBC_GLIBC)
+/*
+ * The following is using imported code from Chromium's
+ * sandbox/linux/services/namespace_sandbox.cc
+ */
+
+#  if !defined(CHECK_EQ)
+#    define CHECK_EQ(a, b) MOZ_RELEASE_ASSERT((a) == (b))
+#  endif
+
+// for sys_gettid()
+using namespace sandbox;
+
+#  include "glibc_hack/namespace_sandbox.inc"
+#endif  // defined(LIBC_GLIBC)
+
 // The libc clone() routine insists on calling a provided function on
 // a new stack, even if the address space isn't shared and it would be
 // safe to expose the underlying system call's fork()-like behavior.
@@ -406,13 +510,14 @@ static int CloneCallee(void* aPtr) {
 // Valgrind would disapprove of using clone() without CLONE_VM;
 // Chromium uses the raw syscall as a workaround in that case, but
 // we don't currently support sandboxing under valgrind.
-MOZ_NEVER_INLINE MOZ_ASAN_BLACKLIST static pid_t DoClone(int aFlags,
-                                                         jmp_buf* aCtx) {
-  uint8_t miniStack[PTHREAD_STACK_MIN];
+MOZ_NEVER_INLINE MOZ_ASAN_IGNORE static pid_t DoClone(int aFlags,
+                                                      jmp_buf* aCtx) {
+  static constexpr size_t kStackAlignment = 16;
+  uint8_t miniStack[4096] __attribute__((aligned(kStackAlignment)));
 #ifdef __hppa__
   void* stackPtr = miniStack;
 #else
-  void* stackPtr = ArrayEnd(miniStack);
+  void* stackPtr = std::end(miniStack);
 #endif
   return clone(CloneCallee, stackPtr, aFlags, aCtx);
 }
@@ -429,15 +534,30 @@ static pid_t ForkWithFlags(int aFlags) {
                                CLONE_CHILD_CLEARTID;
   MOZ_RELEASE_ASSERT((aFlags & kBadFlags) == 0);
 
+  // Block signals due to small stack in DoClone.
+  sigset_t oldSigs;
+  BlockAllSignals(&oldSigs);
+
+  int ret = 0;
   jmp_buf ctx;
   if (setjmp(ctx) == 0) {
     // In the parent and just called setjmp:
-    return DoClone(aFlags | SIGCHLD, &ctx);
+    ret = DoClone(aFlags | SIGCHLD, &ctx);
+    // ret is >0 on success (a valid tid) or -1 on error
+    MOZ_DIAGNOSTIC_ASSERT(ret != 0);
   }
-  // In the child and have longjmp'ed:
-  return 0;
+  // The child longjmps to here, with ret = 0.
+  RestoreSignals(&oldSigs);
+#if defined(LIBC_GLIBC)
+  if (ret == 0) {
+    MaybeUpdateGlibcTidCache();
+  }
+#endif
+  return ret;
 }
 
+// Returns true for success, or returns false and sets errno on
+// failure.  Intended only for procfs pseudo-files.
 static bool WriteStringToFile(const char* aPath, const char* aStr,
                               const size_t aLen) {
   int fd = open(aPath, O_WRONLY);
@@ -446,6 +566,11 @@ static bool WriteStringToFile(const char* aPath, const char* aStr,
   }
   ssize_t written = write(fd, aStr, aLen);
   if (close(fd) != 0 || written != ssize_t(aLen)) {
+    // procfs shouldn't ever cause a short write, but ensure that
+    // errno is set to something distinctive if it does
+    if (written >= 0) {
+      errno = EMSGSIZE;
+    }
     return false;
   }
   return true;
@@ -464,6 +589,7 @@ static void ConfigureUserNamespace(uid_t uid, gid_t gid) {
   len = static_cast<size_t>(SafeSPrintf(buf, "%d %d 1", uid, uid));
   MOZ_RELEASE_ASSERT(len < sizeof(buf));
   if (!WriteStringToFile("/proc/self/uid_map", buf, len)) {
+    SANDBOX_LOG_ERRNO("writing /proc/self/uid_map");
     MOZ_CRASH("Failed to write /proc/self/uid_map");
   }
 
@@ -476,17 +602,18 @@ static void ConfigureUserNamespace(uid_t uid, gid_t gid) {
   len = static_cast<size_t>(SafeSPrintf(buf, "%d %d 1", gid, gid));
   MOZ_RELEASE_ASSERT(len < sizeof(buf));
   if (!WriteStringToFile("/proc/self/gid_map", buf, len)) {
+    SANDBOX_LOG_ERRNO("writing /proc/self/gid_map");
     MOZ_CRASH("Failed to write /proc/self/gid_map");
   }
 }
 
 static void DropAllCaps() {
   if (!LinuxCapabilities().SetCurrent()) {
-    SANDBOX_LOG_ERROR("capset (drop all): %s", strerror(errno));
+    SANDBOX_LOG_ERRNO("capset (drop all)");
   }
 }
 
-pid_t SandboxFork::Fork() {
+pid_t SandboxLaunch::Fork() {
   if (mFlags == 0) {
     MOZ_ASSERT(mChrootServer < 0);
     return fork();
@@ -502,15 +629,29 @@ pid_t SandboxFork::Fork() {
   // can't run atfork hooks.)
   sigset_t oldSigs;
   BlockAllSignals(&oldSigs);
+
+#if defined(MOZ_ENABLE_FORKSERVER)
+  run_moz_pthread_atfork_handlers_prefork();
+#endif
+
   pid_t pid = ForkWithFlags(mFlags);
   if (pid != 0) {
+#if defined(MOZ_ENABLE_FORKSERVER)
+    run_moz_pthread_atfork_handlers_postfork_parent();
+#endif
+
     RestoreSignals(&oldSigs);
     return pid;
   }
 
+#if defined(MOZ_ENABLE_FORKSERVER)
+  run_moz_pthread_atfork_handlers_postfork_child();
+#endif
+
   // WARNING: all code from this point on (and in StartChrootServer)
   // must be async signal safe.  In particular, it cannot do anything
   // that could allocate heap memory or use mutexes.
+  prctl(PR_SET_NAME, "Sandbox Forked");
 
   // Clear signal handlers in the child, under the assumption that any
   // actions they would take (running the crash reporter, manipulating
@@ -523,14 +664,21 @@ pid_t SandboxFork::Fork() {
     StartChrootServer();
   }
 
-  // execve() will drop capabilities, but it seems best to also drop
-  // them here in case they'd do something unexpected in the generic
-  // post-fork code.
+  // execve() will drop capabilities, but the fork server case doesn't
+  // exec so we need to do this directly.  (Also, it's a good idea to
+  // follow the principle of least privilege even when not strictly
+  // necessary.)
+  //
+  // Note that, while capabilities within an unprivileged user
+  // namespace are constrained in theory, in practice they expose a
+  // lot of attack surface and there have been exploitable kernel bugs
+  // related to that in the past, so we really want to drop them
+  // before doing anything that needs sandboxing.
   DropAllCaps();
   return 0;
 }
 
-void SandboxFork::StartChrootServer() {
+void SandboxLaunch::StartChrootServer() {
   // Run the rest of this function in a separate process that can
   // chroot() on behalf of this process after it's sandboxed.
   pid_t pid = ForkWithFlags(CLONE_FS);
@@ -540,12 +688,13 @@ void SandboxFork::StartChrootServer() {
   if (pid > 0) {
     return;
   }
+  prctl(PR_SET_NAME, "Chroot Helper");
 
   LinuxCapabilities caps;
   caps.Effective(CAP_SYS_CHROOT) = true;
   if (!caps.SetCurrent()) {
-    SANDBOX_LOG_ERROR("capset (chroot helper): %s", strerror(errno));
-    MOZ_DIAGNOSTIC_ASSERT(false);
+    SANDBOX_LOG_ERRNO("capset (chroot helper)");
+    MOZ_DIAGNOSTIC_CRASH("caps.SetCurrent() failed");
   }
 
   base::CloseSuperfluousFds(this, [](void* aCtx, int aFd) {
@@ -554,6 +703,9 @@ void SandboxFork::StartChrootServer() {
 
   char msg;
   ssize_t msgLen = HANDLE_EINTR(read(mChrootServer, &msg, 1));
+  if (msgLen < 0) {
+    SANDBOX_LOG_ERRNO("chroot server couldn't read request");
+  }
   if (msgLen == 0) {
     // Process exited before chrooting (or chose not to chroot?).
     _exit(0);
@@ -566,7 +718,10 @@ void SandboxFork::StartChrootServer() {
   // exits at the end of this function, and which is always
   // unwriteable.
   int rv = chroot("/proc/self/fdinfo");
-  MOZ_RELEASE_ASSERT(rv == 0);
+  if (rv != 0) {
+    SANDBOX_LOG_ERRNO("chroot");
+    MOZ_CRASH("chroot failed");
+  }
 
   // Drop CAP_SYS_CHROOT ASAP.  This must happen before responding;
   // the main child won't be able to waitpid(), so it could start
@@ -577,10 +732,16 @@ void SandboxFork::StartChrootServer() {
   // remove that.  (Note: if the process can obtain directory fds, for
   // example via SandboxBroker, it must be blocked from using fchdir.)
   rv = chdir("/");
-  MOZ_RELEASE_ASSERT(rv == 0);
+  if (rv != 0) {
+    SANDBOX_LOG_ERRNO("chdir(\"/\")");
+    MOZ_CRASH("chdir(\"/\") failed");
+  }
 
   msg = kSandboxChrootResponse;
   msgLen = HANDLE_EINTR(write(mChrootServer, &msg, 1));
+  if (msgLen < 0) {
+    SANDBOX_LOG_ERRNO("chroot server couldn't send response");
+  }
   MOZ_RELEASE_ASSERT(msgLen == 1);
   _exit(0);
 }
